﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Threading;
using System.IO;

namespace BaiduRtcDemoCsp
{
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi, Pack = 1)]
    public struct RTC_CONFIG {
        public bool isOnlyPullStream;
        public Int16 h264GopSize;
        public bool isFrameDynamicBind;
        public bool isEnableMultistream;
    }
    [StructLayout(LayoutKind.Sequential)]
    public struct RTC_EVENT_DATA
    {
        public long userId;
        public long streamId;
        public long erroCodes;
        public IntPtr message;
        public IntPtr reason;
    }
    // 带宽信息统计
    [StructLayout(LayoutKind.Sequential)]
    public struct RTC_EVENT_BW_DATA
    {
        // 上行网络可用带宽
        public int upNetAvailableBW;
        // 上行网络已占用带宽
        public int upNetUsedBW;
        // 下行网络可用带宽
        public int downNetAvailableBW;
        // 下行网络已占用带宽
        public int downUsedBW;
    }
	// rtc 房间事件结构体, eventType 参考 BaiduRtcCommonDefine.h
    [StructLayout(LayoutKind.Sequential)]
	public struct RTC_ROOM_EVENT {
		public int eventType;
		public long data;
        [MarshalAs(UnmanagedType.Struct)]
        public RTC_EVENT_DATA eventData;
        [MarshalAs(UnmanagedType.Struct)]
        public RTC_EVENT_BW_DATA bw;
	};

    public enum RtcLiveTransferMode {
        RTC_LIVE_TRANSFER_MODE_ROOM_TRANSMISSION = 0,
        RTC_LIVE_TRANSFER_MODE_ANCHOR_TRANSMISSION = 1
    }

    public enum CaptureType
    {
        kScreen = 0,
        kWindow = 1,
    }

    public struct WindowSource {
        public long id;
        public IntPtr title;
    }

    public enum ShareScreenMode
    {
        kOnlyScreen = 0,
        kWithPreviewVideo,
    }

    public struct UserInfo
    {
        public long userId;
        public String display;
    }

    public enum RtcRoomMode 
    {
        kNormalRoom,
        kLargeRoom,
    }

    //msg show delegate
    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void ShowArrivalMessage(string msg);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void DisplayLogMsg(string msg);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void DllMsgArrivalCallBck(String msg);

	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
	public delegate void DllEventArrivalCallBck(IntPtr events);

	[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void DllVideoArrivalCallBck([MarshalAs(UnmanagedType.U1)] ref byte videoData, int width, int height);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void DisplayVideo(System.Drawing.Bitmap bitmap);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void DisplayRemoteVideo(System.Drawing.Bitmap bitmap, int index);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void RetryDelegate();

    public partial class unSubscribeBtn : Form
    {
        [DllImport(@"RtcWindowsSdkDll.dll",
            EntryPoint = "brtc_create_client",
            CharSet = CharSet.Ansi,
            ExactSpelling = false,
            CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr brtc_create_client();

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_rtc_config",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr brtc_set_rtc_config(IntPtr rtc_client, RTC_CONFIG config);

        [DllImport(@"RtcWindowsSdkDll.dll",
            EntryPoint = "brtc_destroy_client",
            CharSet = CharSet.Ansi,
            ExactSpelling = false,
            CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_destroy_client(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
            EntryPoint = "brtc_init_sdk",
            CharSet = CharSet.Ansi,
            ExactSpelling = false,
            CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_init_sdk(IntPtr rtc_client, String app_id, String token, String cer_path);

        [DllImport(@"RtcWindowsSdkDll.dll",
          EntryPoint = "brtc_deInit_sdk",
          CharSet = CharSet.Ansi,
          ExactSpelling = false,
          CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_deInit_sdk(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
         EntryPoint = "brtc_login_room",
         CharSet = CharSet.Ansi,
         ExactSpelling = false,
         CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_login_room(IntPtr rtc_client, String room_name, String user_id, String token);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_logout_room",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_logout_room(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_publish_streaming",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_publish_streaming(IntPtr rtc_client, Boolean is_auto_subscribe_stream);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_publish_streaming_with_stream_id",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_publish_streaming_with_stream_id(IntPtr rtc_client, String[] streamingIds, int streamingNum);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_stop_publish_streaming",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_stop_publish_streaming(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_subscribe_streaming",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_subscribe_streaming(IntPtr rtc_client, String stream_id);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_stop_subscribe_streaming",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_stop_subscribe_streaming(IntPtr rtc_client, String stream_id);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_only_publish_audio",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_only_publish_audio(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_only_publish_video",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_only_publish_video(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_mute_camera",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_mute_camera(IntPtr rtc_client, Boolean is_mute);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_mute_micphone",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_mute_micphone(IntPtr rtc_client, Boolean is_mute);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_send_message",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_send_message(IntPtr rtc_client, String message);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_send_message_to_user",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_send_message_to_user(IntPtr rtc_client, String message, long userId);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_message_arrival_callback",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_message_arrival_callback(IntPtr rtc_client, DllMsgArrivalCallBck msgArrivalCallBck);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_event_arrival_callback",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_event_arrival_callback(IntPtr rtc_client, DllEventArrivalCallBck evetArrivalCallBck);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_video_arrival_callback",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_video_arrival_callback(IntPtr rtc_client, DllVideoArrivalCallBck videoArrivalCallBck, Int64 observerId);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_sdk_version",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr brtc_get_sdk_version(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_audio_record_device_num",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern int brtc_get_audio_record_device_num(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_audio_record_devices",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_get_audio_record_devices(IntPtr rtc_client,
                                                                  ref byte nameList, int listLen, int maxNameLen);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_auido_record_device",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_auido_record_device(IntPtr rtc_client, String devName);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_audio_playout_device_num",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern int brtc_get_audio_playout_device_num(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_audio_playout_devices",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_get_audio_playout_devices(IntPtr rtc_client,
                                                                  ref byte nameList, int listLen, int maxNameLen);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_auido_playout_device",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_auido_playout_device(IntPtr rtc_client, String devName);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_enable_rtc_stats_info_dot",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_enable_rtc_stats_info_dot(IntPtr rtc_client, Boolean isEnable);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_enable_rtc_logger",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_enable_rtc_logger(IntPtr rtc_client, Boolean isEnable);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_room_id",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern System.UInt64 brtc_get_room_id(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_start_preview",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_start_preview(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_stop_preview",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_stop_preview(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_start_share_screen",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_start_share_screen(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_start_share_screen_with_source",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_start_share_screen_with_source(IntPtr rtc_client, long sourceId);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_stop_share_screen",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_stop_share_screen(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_window_source_list",
        CharSet = CharSet.Unicode,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr brtc_get_window_source_list(IntPtr rtc_client, CaptureType type, ref int length);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_free_window_source_list",
        CharSet = CharSet.Unicode,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_free_window_source_list(IntPtr rtc_client, IntPtr source_list, int length);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_start_share_window",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_start_share_window(IntPtr rtc_client, long sourceId);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_stop_share_window",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_stop_share_window(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_config_live_stream_with_url",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_config_live_stream_with_url(IntPtr rtc_client, String url, Boolean is_mix, Boolean is_recording,
            String mix_template, RtcLiveTransferMode transfer_mode);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_start_live_server_streaming",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_start_live_server_streaming(IntPtr rtc_client, String url, Boolean is_mix, Boolean is_recording,
            String mix_template, RtcLiveTransferMode transfer_mode);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_stop_live_server_streaming",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_stop_live_server_streaming(IntPtr rtc_client, RtcLiveTransferMode transfer_mode);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_video_resolution",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_set_video_resolution(IntPtr rtc_client, int width, int height);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_video_bitrate",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_set_video_bitrate(IntPtr rtc_client, int bit_rate);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_share_screen_mode",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_set_share_screen_mode(IntPtr rtc_client, ShareScreenMode type);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_preview_video_arrival_callback",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern Boolean brtc_set_preview_video_arrival_callback(IntPtr rtc_client, DllVideoArrivalCallBck videoArrivalCallBck, Int64 observerId);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_server_url",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_server_url(IntPtr rtc_client, String serverUrl);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_start_mic_recording",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_start_mic_recording(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_enable_loopback_recording",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_enable_loopback_recording(IntPtr rtc_client, bool enabled);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_loopback_recording_volume",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_loopback_recording_volume(IntPtr rtc_client, int volume);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_loopback_recording_mute",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_loopback_recording_mute(IntPtr rtc_client, bool mute);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_only_screen_share",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_only_screen_share(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_room_mode",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_room_mode(IntPtr rtc_client, RtcRoomMode mode);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_get_remote_audio_levels",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr brtc_get_remote_audio_levels(IntPtr rtc_client);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_destory_remote_audio_levels",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_destory_remote_audio_levels(IntPtr rtc_client, IntPtr audioLevels);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_auto_subscribe",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_auto_subscribe(IntPtr rtc_client, Boolean autoSubscribe);

        [DllImport(@"RtcWindowsSdkDll.dll",
        EntryPoint = "brtc_set_auto_publish",
        CharSet = CharSet.Ansi,
        ExactSpelling = false,
        CallingConvention = CallingConvention.Cdecl)]
        public static extern void brtc_set_auto_publish(IntPtr rtc_client, Boolean autoPublish);


        bool g_switch = false;
        bool g_is_exit = false;
        IntPtr brtc_client = IntPtr.Zero;
        private DllMsgArrivalCallBck msg_arrival_cb;
        private DllEventArrivalCallBck event_arrival_cb;
        private DllVideoArrivalCallBck video_arrival_cb;
        private DllVideoArrivalCallBck remote_video_arrival_cb;
        private DllVideoArrivalCallBck preview_video_arrival_cb;
        private ShowArrivalMessage sam;
        private DisplayLogMsg dlm;
        private DisplayVideo dv;
        private DisplayVideo drv;
        private DisplayRemoteVideo drRemoteV;
        private RetryDelegate retryDelegate;

        private int localVideoWidth = 0;
        private int localVideoHeight = 0;
        private byte[] localVideoBuff;

        private int remoteVideoWidth = 0;
        private int remoteVideoHeight = 0;
        private byte[] remoteVideoBuff;

        private string recordDevName;
        private string playoutDevName;

        //Binding dynamically video frame and observer.
        private bool isDynamicBindFrameAndObserver = true;
        private bool isAutoSubscribe = true;        

        private bool isShareScreen = false;
        private bool isShareSourceScreen = false;
        private bool isStartLiveStream = false;
        private bool isStartPreivew = false;
        private bool isStartLoopbackRecord = false;

        private System.Windows.Forms.Timer retryTimers = new System.Windows.Forms.Timer();
        private int retryStep = 0;
        private List<UserInfo> userInfos = new List<UserInfo>();
        private List<long> subscribedUserIds = new List<long>();

        private int currentIndex = 0;

        private System.Windows.Forms.Timer pageTimers = new System.Windows.Forms.Timer();

        public unSubscribeBtn()
        {
            InitializeComponent();
        }

        public void showMsg(string msg)
        {
            receiveMsgTxtBox.AppendText(DateTime.Now.ToString("HH:mm:ss ") + msg + "\r\n");
        }

        public void msgArrival(String msg)
        {
            receiveMsgTxtBox.Invoke(sam, msg);
        }

        public void eventArrival(IntPtr pData)
        {
            RTC_ROOM_EVENT roomEvent = new RTC_ROOM_EVENT();
            roomEvent = (RTC_ROOM_EVENT)Marshal.PtrToStructure(pData, typeof(RTC_ROOM_EVENT));
            Console.WriteLine("RTC room event, the type is " + roomEvent.eventType + " and userId: " + roomEvent.data);
            switch (roomEvent.eventType)
            {
                case 200: // 正在建立信令通道
                    displayLog("connecting signal server");
                    break;
                case 201: // 信令通道成功
                    displayLog("connect signal server success.");
                    break;
                case 202: // 信令通道建立失败
                    displayLog("signal server connect fail.");
                    reConnect();
                    break;
                case 203: // 信令通道关闭
                    displayLog("signal server disconnected.");
                    break;
                case 204: // 信令通道Server心跳超时
                    displayLog("signal server heart timeout.");
                    reConnect();
                    break;
                case 205: // 信令通道Client发送超时
                    displayLog("signal server send message timeout.");
                    reConnect();
                    break;
                case 206: // 信令通道收到hangup(消息格式等错误，被服务端强制关闭连接)
                    displayLog("signal server receive hangup.");
                    break;
                case 207: // 自己加入房间成功
                    displayLog("The user self joined successfully room.");
                    break;
                case 208: // 自己加入房间失败
                    displayLog("The user self failed to join room.");
                    reConnect();
                    break;
                case 209:
                    displayLog("The remote user leaved room." + roomEvent.data);
                    int index = -1;
                    for (int i = 0; i < userInfos.Count; i++)
                    {
                        if (userInfos.ElementAt(i).userId == roomEvent.data)
                        {
                            index = i;
                            break;
                        }
                    }
                    if (index > 0)
                    {
                        userInfos.RemoveAt(index);
                    }
                    break;
                case 210:
                    displayLog("The remote user joined room." + roomEvent.data);
                    UserInfo userInfo = new UserInfo();
                    userInfo.userId = roomEvent.data;
                    userInfos.Add(userInfo);
                    if (isDynamicBindFrameAndObserver)
                    {
                        autoSubscribeRemoteUser(roomEvent.data);
                    }
                    break;
                case 214:
                    {
                        displayLog("receive message." + Marshal.PtrToStringAnsi(roomEvent.eventData.message) + " from user " + roomEvent.eventData.userId);
                    }
                    break;
                default:
                    displayLog("The other room events coming ......" + roomEvent.eventType);
                    break;

            }
        }

        public void reConnect()
        {
            //destory
//            logout.BeginInvoke(retryDelegate);
        }

        public void videoArrival(ref byte videoData, int width, int height)
        {
            //displayLog("video coming ....");
            //logOutput.Invoke(dlm, "video data incoming ...... \r\n");
            if (!g_is_exit)
            {
                rgbaTogBitmap(ref videoData, width, height);
            }
        }

        public void remoteVideoArrival(ref byte videoData, int width, int height)
        {
            //Console.Write("remote video arrival ...... \r\n");
            //logOutput.Invoke(dlm, "video data incoming ......");
            if (!g_is_exit)
            {
                rgbaTogBitmapRemote(ref videoData, width, height);
            }
        }

        public void previewVideoArrival(ref byte videoData, int width, int height)
        {
            //Console.Write("remote video arrival ...... \r\n");
            //logOutput.Invoke(dlm, "video data incoming ......");
            if (!g_is_exit)
            {
                rgbaTogBitmapRemote(ref videoData, width, height);
            }
        }

        public void displayVideo(System.Drawing.Bitmap bitmap)
        {
            localVideoPb.Image = bitmap;
            localVideoPb.Image.RotateFlip(RotateFlipType.Rotate180FlipX);
        }

        public void displayRemoteVideo(System.Drawing.Bitmap bitmap)
        {
            remoteVideoBox.Image = bitmap;
            remoteVideoBox.Image.RotateFlip(RotateFlipType.Rotate180FlipX);
        }

        public void displayUserRemoteVideo(System.Drawing.Bitmap bp, int index)
        {
            switch (index)
            {
                case 0:
                    remoteVideoBox.Image = bp;
                    remoteVideoBox.Image.RotateFlip(RotateFlipType.Rotate180FlipX);
                    break;
                case 1:
                    remoteVideoBox1.Image = bp;
                    remoteVideoBox1.Image.RotateFlip(RotateFlipType.Rotate180FlipX);
                    break;
                case 2:
                    remoteVideoBox2.Image = bp;
                    remoteVideoBox2.Image.RotateFlip(RotateFlipType.Rotate180FlipX);
                    break;
                case 3:
                    remoteVideoBox3.Image = bp;
                    remoteVideoBox3.Image.RotateFlip(RotateFlipType.Rotate180FlipX);
                    break;
            }
        }

        private struct UserVideoInfo {
            public int index;
            public int remoteVideoWidth;
            public int remoteVideoHeight;
            public byte[] remoteVideoBuff;
        }

        private Dictionary<long, UserVideoInfo> userVideoInfoMap = new Dictionary<long, UserVideoInfo>();

        private void getRemoteUserVideoBuffer(int width, int height, long userId)
        {
            UserVideoInfo info;
            if (userVideoInfoMap.ContainsKey(userId))
            {
                info = userVideoInfoMap[userId];
            }
            else
            {
                info = new UserVideoInfo();
            }
            if (width != info.remoteVideoWidth || height != info.remoteVideoHeight)
            {
                info.remoteVideoWidth = width;
                info.remoteVideoHeight = height;
                int len = width * height * 4;
                info.remoteVideoBuff = new byte[len];
            }
            userVideoInfoMap[userId] = info;
        }

        private void onUserRemoteVideoArrival(ref byte videoData, int width, int height, long userId)
        {
            int dataLen = width * height * 4;
            getRemoteUserVideoBuffer(width, height, userId);
            UserVideoInfo info = userVideoInfoMap[userId];
            unsafe
            {
                fixed (byte* pImage = &videoData)
                {
                    using (UnmanagedMemoryStream ms = new UnmanagedMemoryStream((byte*)pImage, dataLen, dataLen, FileAccess.Read))
                    {
                        ms.Read(info.remoteVideoBuff, 0, dataLen);
                    }
                }
            }

            Int64 stride = width * 4;
            GCHandle handle = GCHandle.Alloc(info.remoteVideoBuff, GCHandleType.Pinned);
            Int64 scan0 = (Int64)handle.AddrOfPinnedObject();
            scan0 += (height - 1) * stride;
            System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(width, height, -(int)stride, System.Drawing.Imaging.PixelFormat.Format32bppRgb, (IntPtr)scan0);
            handle.Free();

            remoteVideoBox.BeginInvoke(drRemoteV, bitmap, userIndexMap[userId]);
        }

        public void onRetry()
        {
            retryTimers.Stop();
            retryStep = 0;

            retryTimers.Interval = 500;
            retryTimers.Enabled = true;
            retryTimers.Start();
        }

        private void onRetryTick(Object myobject, EventArgs args)
        {
            switch (retryStep) {
                case 0:
                    logout.PerformClick();
                    break;
                case 1:
                    releaseBtn.PerformClick();
                    retryTimers.Stop();
                    retryTimers.Interval = 5000;
                    retryTimers.Start();
                    break;
                case 2:
                    createBtn.PerformClick();
                    retryTimers.Stop();
                    retryTimers.Interval = 500;
                    retryTimers.Start();
                    break;
                case 3:
                    loginBtn.PerformClick();
                    break;
            }
            retryStep++;
            if (retryStep == 4)
            {
                retryTimers.Stop();
            }
        }

        public void getLocalVideoBuff(int width, int height)
        {
            if (width != localVideoWidth || height != localVideoHeight)
            {
                localVideoWidth = width;
                localVideoHeight = height;
                int len = width * height * 4;
                localVideoBuff = new byte[len];
            }
        }

        public void rgbaTogBitmap(ref byte image, int width, int height)
        {
            //Console.WriteLine("Local video width" + width + " " + "height:" + height);
            int dataLen = width * height * 4;
            //if (dataLen <= 0)
            //{
            //    return;
            //}
            //Console.WriteLine("start new byte array");
            //byte[] tmpBuf = new byte[dataLen];
            //Console.WriteLine("end new byte array");
            getLocalVideoBuff(width, height);
            unsafe
            {
                fixed (byte* pImage = &image)
                {
                    using (UnmanagedMemoryStream ms = new UnmanagedMemoryStream((byte*)pImage, dataLen, dataLen, FileAccess.Read))
                    {
                        ms.Read(localVideoBuff, 0, dataLen);
                    }
                }
            }
            Int64 stride = width * 4;
            GCHandle handle = GCHandle.Alloc(localVideoBuff, GCHandleType.Pinned);
            Int64 scan0 = (Int64)handle.AddrOfPinnedObject();
            scan0 += (height - 1) * stride;
            System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(width, height, -(int)stride, System.Drawing.Imaging.PixelFormat.Format32bppRgb, (IntPtr)scan0);
            handle.Free();
            //logOutput.Invoke(dlm, "The video data converted to bitmap from rgba.");
            //display bitmap
            localVideoPb.BeginInvoke(dv, bitmap);
            GC.Collect();
        }

        public void getRemoteVideoBuffer(int width, int height)
        {
            if (width != remoteVideoWidth || height != remoteVideoHeight)
            {
                remoteVideoWidth = width;
                remoteVideoHeight = height;
                int len = width * height * 4;
                remoteVideoBuff = new byte[len];
            }
        }

        public void rgbaTogBitmapRemote(ref byte image, int width, int height)
        {
            //Console.WriteLine("Remote video width" + width + "" + "height:" + height);
            int dataLen = width * height * 4;
            getRemoteVideoBuffer(width, height);
            //if (dataLen <= 0)
            //{
            //    return;
            //}
            //byte[] tmpBuf = new byte[dataLen];
            unsafe
            {
                fixed (byte* pImage = &image)
                {
                    using (UnmanagedMemoryStream ms = new UnmanagedMemoryStream((byte*)pImage, dataLen, dataLen, FileAccess.Read))
                    {
                        ms.Read(remoteVideoBuff, 0, dataLen);
                    }
                }
            }

            Int64 stride = width * 4;
            GCHandle handle = GCHandle.Alloc(remoteVideoBuff, GCHandleType.Pinned);
            Int64 scan0 = (Int64)handle.AddrOfPinnedObject();
            scan0 += (height - 1) * stride;
            System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(width, height, -(int)stride, System.Drawing.Imaging.PixelFormat.Format32bppRgb, (IntPtr)scan0);
            handle.Free();
            //logOutput.Invoke(dlm, "The video data converted to bitmap from rgba.");
            //display bitmap
            remoteVideoBox.BeginInvoke(drv, bitmap);
        }

        private void createClientAndInit(object sender, EventArgs e)
        {
            //judge target platform
            if (IntPtr.Size == 4)
            {
                Console.WriteLine("x86 target platform.");
            }
            else if (IntPtr.Size == 8)
            {
                Console.WriteLine("x64 target platform.");
            }
            else
            {
                Console.WriteLine("anyCput target platform.");
            }

            //CREATE
            displayLog("creating brtc_client.");
            brtc_client = brtc_create_client();
            IntPtr iPtr = brtc_get_sdk_version(brtc_client);
            displayLog(Marshal.PtrToStringAnsi(iPtr));

//            brtc_set_rtc_config(brtc_client, config);
            String app_id = "appjfkryqbzyns5";
            String token = "";
            String cer_path = System.AppDomain.CurrentDomain.BaseDirectory + "a.cer";
            displayLog("init sdk.");
            brtc_init_sdk(brtc_client, app_id, token, cer_path);
            brtc_enable_rtc_logger(brtc_client, true);
            
            // 仅使用屏幕分享
        //    brtc_set_only_screen_share(brtc_client);

            // textbox show message delegate
            sam = new ShowArrivalMessage(showMsg);
            dlm = new DisplayLogMsg(displayLog);
            dv = new DisplayVideo(displayVideo);
            drv = new DisplayVideo(displayRemoteVideo);
            drRemoteV = new DisplayRemoteVideo(displayUserRemoteVideo);

            retryDelegate = new RetryDelegate(onRetry);
            pageTimers.Tick += new EventHandler(subscribeRemoteUser);
            retryTimers.Tick += new EventHandler(onRetryTick);

            //set msg arrival callback
            displayLog("setting message arrival callback");
            msg_arrival_cb = new DllMsgArrivalCallBck(msgArrival);
            brtc_set_message_arrival_callback(brtc_client, msg_arrival_cb);

            //set event arrival callback
            event_arrival_cb = new DllEventArrivalCallBck(eventArrival);
            brtc_set_event_arrival_callback(brtc_client, event_arrival_cb);
            //
            //brtc_set_share_screen_mode(brtc_client, ShareScreenMode.kWithPreviewVideo);

            if (!isDynamicBindFrameAndObserver)
            {
                //set video arrival callback
                //displayLog("setting video arrival callback");
                video_arrival_cb = new DllVideoArrivalCallBck(videoArrival);
                brtc_set_video_arrival_callback(brtc_client, video_arrival_cb, 0);

                //setting remote video arrival callback
                remote_video_arrival_cb = new DllVideoArrivalCallBck(remoteVideoArrival);
                brtc_set_video_arrival_callback(brtc_client, remote_video_arrival_cb, 09098);
            }

            preview_video_arrival_cb = new DllVideoArrivalCallBck(previewVideoArrival);
            brtc_set_preview_video_arrival_callback(brtc_client, preview_video_arrival_cb, 100000);

            //get audio recording device name list
            //int recordNum = brtc_get_audio_record_device_num(brtc_client);
            byte[] nameList = new byte[512];
            brtc_get_audio_record_devices(brtc_client, ref nameList[0], 512, 128);
            byte[] dev1Name = nameList.Skip(0).Take(128).ToArray();
            String str = System.Text.Encoding.GetEncoding("utf-8").GetString(dev1Name);
            //string str = Encoding.UTF8.GetString(dev1Name);
            str = str.Replace("\0", "");
            Console.WriteLine(str);
            recordDevName = str;
            brtc_set_auido_record_device(brtc_client, str);
            //Console.WriteLine(System.Text.Encoding.GetEncoding("GB2312").GetString(nameList));

            // get audio playout devices numbers.
            byte[] playoutDevList = new byte[512];
            int playoutNum = brtc_get_audio_playout_device_num(brtc_client);
            brtc_get_audio_playout_devices(brtc_client, ref playoutDevList[0], 512, 128);
            int j = 0;
            for (int i = 0; i < playoutNum; i++)
            {
                Console.WriteLine("Audio playout device num is " + playoutNum);
                byte[] dev = playoutDevList.Skip(j).Take(128).ToArray();
                String playoutDevName = System.Text.Encoding.GetEncoding("utf-8").GetString(dev);
                //string str = Encoding.UTF8.GetString(dev1Name);
                playoutDevName = playoutDevName.Replace("\0", "");
                Console.WriteLine(playoutDevName);
                if (i == 0)
                {
                    Console.WriteLine("brtc_set_auido_playout_device " + i + " @c#Demo.");
                    brtc_set_auido_playout_device(brtc_client, playoutDevName);
                }
                j += 128;
            }
        }

        private void loginAndPublishStream(object sender, EventArgs e)
        {
            //LOGIN
            Console.WriteLine("Begin to login room!!!!");
            g_is_exit = false;
            if (!g_switch) {
                g_switch = true;
            } else {
                return;
            }

            if (brtc_client != IntPtr.Zero)
            {
                
                String room_name = roomId.Text.Length > 0 ? roomId.Text : "999";
                String user_id = new Random().Next(10000,10000000).ToString();
                String token = "";
                //                brtc_set_server_url(brtc_client, "");
                //lss live stream
                //brtc_config_live_stream_with_url(brtc_client, "", true, false, "sigle_video_with_mix_audio", RtcLiveTransferMode.RTC_LIVE_TRANSFER_MODE_ROOM_TRANSMISSION);
                //                brtc_set_video_resolution(brtc_client, 1920 ,1080);
                //                brtc_set_video_bitrate(brtc_client, 5000000);
                //brtc_set_only_publish_video(brtc_client);
                brtc_set_auto_publish(brtc_client, true);
                brtc_set_auto_subscribe(brtc_client, isAutoSubscribe);
                brtc_login_room(brtc_client, room_name, user_id, token);
                //brtc_set_only_publish_audio(brtc_client);
                //setting local video frame arrival callback, observer id  as 0
                if (isDynamicBindFrameAndObserver)
                {
                    video_arrival_cb = new DllVideoArrivalCallBck(videoArrival);
                    brtc_set_video_arrival_callback(brtc_client, video_arrival_cb, 0);
                }
                //                String[] subscribes = new string[1];
                //                subscribes[0] = "123456789";
                //                brtc_publish_streaming_with_stream_id(brtc_client, subscribes, 1);
                // brtc_publish_streaming(brtc_client, false);
            //    brtc_start_share_screen_with_source(brtc_client, 0);
            }
        }

        private void logoutRoom(object sender, EventArgs e)
        {
            //LOGOUT
            g_is_exit = true;
            if (!g_switch)
            {
                return;
            } else
            {
                g_switch = false;
            }
            if (brtc_client != IntPtr.Zero)
            {
                //brtc_stop_publish_streaming(brtc_client);
                //                Thread thrd = new Thread(new ThreadStart(new ThreadStart(() => {
                subscribedUserIds.Clear();
                pageTimers.Tick -= new EventHandler(subscribeRemoteUser);
                displayLog("logoutRoom");
                brtc_logout_room(brtc_client);
//                })));
//                thrd.IsBackground = true;
//               thrd.Start();
                //brtc_logout_room(brtc_client);
            }
        }

        private void release(object sender, EventArgs e)
        {
            //RELEASE
            displayLog("release");
            brtc_deInit_sdk(brtc_client);
            brtc_destroy_client(brtc_client);
        }

        private void sendMsgClick(object sender, EventArgs e)
        {
            if (brtc_client != IntPtr.Zero)
            {
                displayLog("start to send message!!!");
                String msg = sendMsgTxtBox.Text;
                brtc_send_message(brtc_client, msg);
            }
        }

        private void closeCameraClick(object sender, EventArgs e)
        {
            if (closeCameraBtn.Text == "关闭摄像头")
            {
                closeCameraBtn.Text = "打开摄像头";
                if (brtc_client != IntPtr.Zero)
                {
                    displayLog("close camera.");
                    brtc_mute_camera(brtc_client, false);
                }

            } else
            {
                closeCameraBtn.Text = "关闭摄像头";
                if (brtc_client != IntPtr.Zero)
                {
                    displayLog("open camera.");
                    brtc_mute_camera(brtc_client, true);
                }
            }
        }

        private void closeMicClick(object sender, EventArgs e)
        {
            if (closeMicBtn.Text == "关闭麦克风")
            {
                closeMicBtn.Text = "打开麦克风";
                if (brtc_client != IntPtr.Zero)
                {
                    displayLog("close micphone");
                    brtc_mute_micphone(brtc_client, false);
                }

            } else
            {
                closeMicBtn.Text = "关闭麦克风";
                if (brtc_client != IntPtr.Zero)
                {
                    displayLog("open micphone");
                    brtc_mute_micphone(brtc_client, true);
                }
            }
        }

        public void displayLog(string log)
        {
            if (logOutput.InvokeRequired)
            {
                this.BeginInvoke(dlm, DateTime.Now.ToString("HH:mm:ss ") + log + "\r\n");
            }
            else
            {
                logOutput.AppendText(DateTime.Now.ToString("HH:mm:ss ") + log + "\r\n");
            }
        }

        private void enableRtcLog(object sender, EventArgs e)
        {
            Button button = sender as Button;
            if (string.Equals(button.Text, "open log"))
            {
                button.Text = "close log";
                displayLog("Openning log system to save log as file under exe.");
                if (brtc_client != IntPtr.Zero)
                {
                    brtc_enable_rtc_logger(brtc_client, true);
                }
            } else {
                button.Text = "open log";
                displayLog("Closing log system.");
                if (brtc_client != IntPtr.Zero)
                {
                    brtc_enable_rtc_logger(brtc_client, false);
                }
            }
        }

        private void dotButton_Click(object sender, EventArgs e)
        {
            Button btn = sender as Button;
            if (string.Equals(btn.Text, "open dot")) {
                btn.Text = "close dot";
                displayLog("Openning rtc stats info statistics.");
                if (brtc_client != IntPtr.Zero)
                {
                    brtc_enable_rtc_stats_info_dot(brtc_client, true);
                }
            } else {
                btn.Text = "open dot";
                displayLog("Closing rtc stats info statistics.");
                if (brtc_client != IntPtr.Zero)
                {
                    brtc_enable_rtc_stats_info_dot(brtc_client, false);
                }
            }
        }

        private void roomIdButton_Click(object sender, EventArgs e)
        {
            UInt64 roomId = 0;
            if (brtc_client != IntPtr.Zero)
            {
                roomId = brtc_get_room_id(brtc_client);
            }
            displayLog("Rtc room is " + roomId);
        }


        private void BadiduRtcDemo_Load(object sender, EventArgs e)
        {
        }

        private void screenShareBtn_Click(object sender, EventArgs e)
        {
            if (brtc_client != IntPtr.Zero)
            {
                if (isShareScreen)
                {
                    brtc_stop_share_screen(brtc_client);
                    screenShareBtn.Text = "屏幕共享";
                }
                else
                {
                    brtc_start_share_screen(brtc_client);
                    screenShareBtn.Text = "关闭屏幕共享";
                }
                isShareScreen = !isShareScreen;
            }
        }

        private void startLiveStreamButton_Click(object sender, EventArgs e)
        {
            if (brtc_client != IntPtr.Zero)
            {
                if (isStartLiveStream)
                {
                    brtc_stop_live_server_streaming(brtc_client, RtcLiveTransferMode.RTC_LIVE_TRANSFER_MODE_ROOM_TRANSMISSION);
                    startLiveStreamButton.Text = "开始转推";
                }
                else
                {
                    brtc_start_live_server_streaming(brtc_client, "", true, false, "sigle_video_with_mix_audio", RtcLiveTransferMode.RTC_LIVE_TRANSFER_MODE_ROOM_TRANSMISSION);
                    startLiveStreamButton.Text = "停止转推";
                }
                isStartLiveStream = !isStartLiveStream;
            }
        }

        private void shareScreenWithSourceButton_Click(object sender, EventArgs e)
        {
            if (brtc_client != IntPtr.Zero)
            {
                if (isShareSourceScreen)
                {
                    brtc_stop_share_screen(brtc_client);
                    shareScreenWithSourceButton.Text = "指定屏幕共享";
                }
                else
                {

                    WindowSource[] source = getSourceList(CaptureType.kScreen);
                    int count = Screen.AllScreens.Count();
                    brtc_start_share_screen_with_source(brtc_client, source[0].id);
                    shareScreenWithSourceButton.Text = "关闭指定屏幕共享";
                }
                isShareSourceScreen = !isShareSourceScreen;
            }
        }

        private WindowSource[] getSourceList(CaptureType type)
        {
            int sourceLength = 0;
            IntPtr sourceList = brtc_get_window_source_list(brtc_client, type, ref sourceLength);
            WindowSource[] sourceArray = new WindowSource[sourceLength];
            if (sourceLength > 0)
            {
                var structSize = Marshal.SizeOf(typeof(WindowSource));
                for (int i = 0; i < sourceLength; i++)
                {
                    IntPtr ins = new IntPtr(sourceList.ToInt64() + i * structSize);
                    sourceArray[i] = Marshal.PtrToStructure<WindowSource>(ins);
                }
                for (int i = 0; i < sourceArray.Length; i++)
                {
                    string title = Marshal.PtrToStringUni(sourceArray[i].title);
                }
            }
            brtc_free_window_source_list(brtc_client, sourceList, sourceLength);
            return sourceArray;
        }

        private void windowShareButton_Click(object sender, EventArgs e)
        {
            if (brtc_client != IntPtr.Zero)
            {
                if (isShareScreen)
                {
                    brtc_stop_share_screen(brtc_client);
                    screenShareBtn.Text = "窗口共享";
                }
                else
                {
                    WindowSource[] source = getSourceList(CaptureType.kWindow);
                    brtc_start_share_window(brtc_client, source[0].id);
                    screenShareBtn.Text = "关闭窗口共享";
                }
                isShareScreen = !isShareScreen;
            }
        }

        private void screenShotBtn_Click(object sender, EventArgs e)
        {
            remoteVideoBox.Image.Save("", System.Drawing.Imaging.ImageFormat.Jpeg);
        }

        private void startPreviewBtn_Click(object sender, EventArgs e)
        {
            if (!isStartPreivew)
            {
                brtc_start_preview(brtc_client);
            }
            else
            {
                brtc_stop_preview(brtc_client);
            }
        }

        private void insertMicBtn_Click(object sender, EventArgs e)
        {
            // 重插入麦克风后
            //get audio recording device name list
            //int recordNum = brtc_get_audio_record_device_num(brtc_client);
            byte[] nameList = new byte[512];
            brtc_get_audio_record_devices(brtc_client, ref nameList[0], 512, 128);
            byte[] dev1Name = nameList.Skip(0).Take(128).ToArray();
            String str = System.Text.Encoding.GetEncoding("utf-8").GetString(dev1Name);
            //string str = Encoding.UTF8.GetString(dev1Name);
            str = str.Replace("\0", "");
            Console.WriteLine(str);
            recordDevName = str;
            brtc_set_auido_record_device(brtc_client, str);

            brtc_start_mic_recording(brtc_client);
        }

        private void shareLoopback_Click(object sender, EventArgs e)
        {
            if (isStartLoopbackRecord)
            {
                brtc_enable_loopback_recording(brtc_client, false);
                shareLoopback.Text = "开启桌面音频共享";
                isStartLoopbackRecord = false;
            }
            else
            {
                brtc_enable_loopback_recording(brtc_client, true);
                shareLoopback.Text = "关闭桌面音频共享";
                isStartLoopbackRecord = true;
            }
        }

        private void volumeBar_ValueChanged(object sender, EventArgs e)
        {
            brtc_set_loopback_recording_volume(brtc_client, volumeBar.Value);
        }

        private List<DisplayRemoteVideo> displayList = new List<DisplayRemoteVideo>();
        private void showRemoteVideo(ref byte image, int width, int height, long userId, int index)
        {
            int dataLen = width * height * 4;
            getRemoteVideoBuffer(width, height);
            unsafe
            {
                fixed (byte* pImage = &image)
                {
                    using (UnmanagedMemoryStream ms = new UnmanagedMemoryStream((byte*)pImage, dataLen, dataLen, FileAccess.Read))
                    {
                        ms.Read(remoteVideoBuff, 0, dataLen);
                    }
                }
            }

            Int64 stride = width * 4;
            GCHandle handle = GCHandle.Alloc(remoteVideoBuff, GCHandleType.Pinned);
            Int64 scan0 = (Int64)handle.AddrOfPinnedObject();
            scan0 += (height - 1) * stride;
            System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(width, height, -(int)stride, System.Drawing.Imaging.PixelFormat.Format32bppRgb, (IntPtr)scan0);
            handle.Free();

            remoteVideoBox.BeginInvoke(drRemoteV, bitmap, index);
        }

        private List<DllVideoArrivalCallBck> arrvalList = new List<DllVideoArrivalCallBck>();
        private Dictionary<long, int> userIndexMap = new Dictionary<long, int>();


        private void changePage_Click(object sender, EventArgs e)
        {
            if (subscribedUserIds.Count > 0)
            {
                foreach (long userId in subscribedUserIds)
                {
                    brtc_stop_subscribe_streaming(brtc_client, userId.ToString());
                }
            }

            subscribedUserIds.Clear();
            userIndexMap.Clear();
            arrvalList.Clear();

            pageTimers.Interval = 2000;
            pageTimers.Enabled = true;
            pageTimers.Start();
        }

        private List<int> windowIndexList = new List<int> { 0,0,0,0 };

        private int getUsableWindowIndex()
        {
            for (int i = 0;i < windowIndexList.Count; i++)
            {
                if (windowIndexList[i] == 0)
                {
                    windowIndexList[i] = 1;
                    return i;
                }
            }
            return -1;
        }

        private void autoSubscribeOnLeave(long userId)
        {
            brtc_stop_subscribe_streaming(brtc_client, userId.ToString());
        }

        private void autoSubscribeRemoteUser(long userId)
        {
            long showUserId = userId;
            if (!userIndexMap.ContainsKey(userId))
            {
                int winIndex = getUsableWindowIndex();
                if (winIndex == -1)
                {
                    return;
                }
                userIndexMap[showUserId] = winIndex;
            }
            displayLog("subscribing .... userId:" + showUserId + "  showWindow:" + userIndexMap[showUserId]);
            DllVideoArrivalCallBck callback = (ref byte videoData, int width, int height) => {
                onUserRemoteVideoArrival(ref videoData, width, height, showUserId);
            };
            arrvalList.Add(callback);
            brtc_set_video_arrival_callback(brtc_client, callback, showUserId);
            if (!isAutoSubscribe) {
                brtc_subscribe_streaming(brtc_client, showUserId.ToString());
            }
        }

        private void subscribeRemoteUser(Object myobject, EventArgs args)
        {
            pageTimers.Stop();
            int MaxPageNum = 4;
            if (userInfos.Count > MaxPageNum)
            {
                for (int i = 0; i < MaxPageNum; i++)
                {
                    long userId = userInfos.ElementAt((currentIndex + i) % userInfos.Count).userId;
                    if (!subscribedUserIds.Contains(userId))
                    {
                        subscribedUserIds.Add(userId);
                    }
                }
                currentIndex += currentIndex + MaxPageNum;
            }
            else
            {
                foreach (UserInfo info in userInfos)
                {
                    if (!subscribedUserIds.Contains(info.userId))
                    {
                        subscribedUserIds.Add(info.userId);
                    }
                }
            }

            for (int i = 0; i < subscribedUserIds.Count; i++)
            {
                long showUserId = subscribedUserIds[i];
                userIndexMap[showUserId] = i;
                displayLog("subscribing .... userId:" + showUserId + "  showWindow:" + i);
                DllVideoArrivalCallBck callback = (ref byte videoData, int width, int height) => {
                    onUserRemoteVideoArrival(ref videoData, width, height, showUserId);
                };
                arrvalList.Add(callback);
                brtc_set_video_arrival_callback(brtc_client, callback, showUserId);
                brtc_subscribe_streaming(brtc_client, showUserId.ToString());
            }
        }

        private void sendMessageBtn_Click(object sender, EventArgs e)
        {
            IntPtr levelsPointer = brtc_get_remote_audio_levels(brtc_client);
        }

        private void subscribeBtn_Click(object sender, EventArgs e) {
            brtc_subscribe_streaming(brtc_client, subscribeIdEdit.Text);            
        }

        private void unSubscribeButton_Click(object sender, EventArgs e) {
            brtc_stop_subscribe_streaming(brtc_client, unSubscribeIdEdit.Text);
        }
    }
}
